/* -*- Mode: C++ ; Coding: euc-japan -*- */
/* Time-stamp: <2013-05-14 17:23:23 cyamauch> */

#ifndef _SLI__ASARRAY_H
#define _SLI__ASARRAY_H 1

/**
 * @file   asarray.h
 * @brief  Ǥե饹Υ֥Ϣ򰷤 asarray 饹Υ
 */

#include "sli_config.h"
#include "tarray_tstring.h"
#include "tarray.h"
#include "ctindex.h"

namespace sli
{

/*
 * sli::asarray template class can handle the associative arrays of arbitrary
 * data types or classes.  Its function is almost the same as that of 
 * asarray_tstring, except that this class does not include member functions
 * specializing in string.
 */

/**
 * @class  sli::asarray
 * @brief  Ǥե饹Υ֥Ϣ򰷤Υƥץ졼ȥ饹
 *
 *   asarray ƥץ졼ȥ饹ϡǤե饹Υ֥Ϣ򰷤
 *   Ǥޤ
 *
 * @note  asarray ֥ǤϡͤȤƤΥ֥ȤñȤ
 *        ¸ޤƱˡʸֹȤδط¸졤
 *        ˤ ctindex 饹ѤƤޤctindex 饹ϥʸ
 *        Ȥδط򡤡ڡ׹¤ǵϿ®ʸ(ʸ󤫤ֹ
 *        Ѵ)ǽˤƤޤڡ׹¤ǤΤǡˤϤ
 *        ֤ޤ
 * @author Chisato YAMAUCHI
 * @date 2013-04-01 00:00:00
 */

  template <class datatype> class asarray
  {

  public:
    /* constructor and destructor */
    asarray();
    asarray(const asarray &obj);
    virtual ~asarray();

    /*
     * operator
     */
    /* same as .init(sobj) */
    virtual asarray &operator=(const asarray &obj);

    /* same as .at(key) */
    virtual datatype &operator[]( const char *key );
    virtual const datatype &operator[]( const char *key ) const;

    /*
     * member functions
     */
    /* complete initialization of objects */
    virtual asarray &init();
    virtual asarray &init(const asarray &obj);

    /* padding of the total existing values in an a ssociated array */
    virtual asarray &clean(const datatype &one);

    /* set an associative array */
    virtual asarray &assign( const asarray &src );
    virtual asarray &assign( const char *key, const datatype &val );

    /* assign multiple strings or arrays of strings to the key */
    virtual asarray &assign_keys( const char *key0, ... );
    virtual asarray &vassign_keys( const char *key0, va_list ap );
    virtual asarray &assign_keys( const char *const *keys );
    virtual asarray &assign_keys( const tarray_tstring &keys );

    /* split a string and assign to the key */
    virtual asarray &split_keys( const char *src_str, const char *delims, 
				 bool zero_str, const char *quotations,
				 int escape, bool rm_escape );
    virtual asarray &split_keys( const char *src_str, const char *delims,
				 bool zero_str = false );
    virtual asarray &split_keys( const tstring &src_str, const char *delims, 
				 bool zero_str, const char *quotations,
				 int escape, bool rm_escape );
    virtual asarray &split_keys( const tstring &src_str, const char *delims,
				 bool zero_str = false );

    /* append an element */
    virtual asarray &append( const asarray &src );
    virtual asarray &append( const char *key, const datatype &val );

    /* insert an element */
    virtual asarray &insert( const char *key, const asarray &src );
    virtual asarray &insert( const char *key,
			     const char *newkey, const datatype &newval );

    /* erase element(s) */
    virtual asarray &erase();
    virtual asarray &erase( const char *key, size_t num_elements = 1 );

    /* swap self and another */
    virtual asarray &swap( asarray &sobj );

    /* change a key string */
    virtual asarray &rename_a_key( const char *org_key, const char *new_key );

    /* a reference to the element value corresponding to specified key or */
    /* element number                                                     */
    virtual datatype &at( const char *key );
    virtual datatype &atf( const char *fmt, ... );
    virtual datatype &vatf( const char *fmt, va_list ap );
#ifdef SLI__OVERLOAD_CONST_AT
    virtual const datatype &at( const char *key ) const;
    virtual const datatype &atf( const char *fmt, ... ) const;
    virtual const datatype &vatf( const char *fmt, va_list ap ) const;
#endif
    virtual const datatype &at_cs( const char *key ) const;
    virtual const datatype &atf_cs( const char *fmt, ... ) const;
    virtual const datatype &vatf_cs( const char *fmt, va_list ap ) const;

    /* length of an associated array */
    virtual size_t length() const;
    /* not recommended */
    virtual size_t size() const;

    /* acquire the element number corresponding to the key string */
    virtual ssize_t index( const char *key ) const;
    virtual ssize_t indexf( const char *fmt, ... ) const;
    virtual ssize_t vindexf( const char *fmt, va_list ap ) const;

    /* acquire the key string corresponding to the element number */
    virtual const char *key( size_t index ) const;

    /* a reference to the element value corresponding to specified key or */
    /* element number                                                     */
    virtual datatype &at( size_t index );
    virtual const datatype &at_cs( size_t index ) const;

    /* a reference to the array object of the key string (read only) */
    virtual const tarray_tstring &keys() const;

    /* a reference to the array object of the value (read only) */
    virtual const tarray<datatype> &values() const;

    /* register a function to create object of datatype class */
    virtual void register_creator( datatype *(*func)(void *), void *user_ptr );

    /* ֥Ȥ return ľ˻Ȥshallow copy Ĥ */
    /* Ȥ (̤)                                                         */
    virtual void set_scopy_flag();

#if 0	/* ȤϤǤʤ餷orz */
    virtual asarray &assign( const char *key0, const datatype &val0,
			     const char *key1, ... );
    virtual asarray &vassign( const char *key0, const datatype &val0,
			      const char *key1, va_list ap );
    virtual asarray &append( const char *key0, const datatype &val0, 
			     const char *key1, ... );
    virtual asarray &vappend( const char *key0, const datatype &val0, 
			      const char *key1, va_list ap );
    virtual asarray &insert( const char *key, 
			     const char *key0, const datatype &val0, 
			     const char *key1, ... );
    virtual asarray &vinsert( const char *key,
			      const char *key0, const datatype &val0, 
			      const char *key1, va_list ap );
#endif

  private:
    tarray_tstring key_rec;
    tarray<datatype> value_rec;
    ctindex index_rec;

    /* for error handling */
    void err_throw1( const char *func0, const char *level0, 
		     const char *message0, const char *arg ) const;
    void err_report1( const char *func0, const char *level0, 
		      const char *message0, const char *arg ) const;

  };
}

namespace sli
{

/* constructor */
/**
 * @brief  󥹥ȥ饯
 */
template <class datatype>
asarray<datatype>::asarray()
{
    return;
}

/* copy constructor */

/**
 * @brief  ԡ󥹥ȥ饯
 *
 *  ȤobjƤǽޤ
 * 
 * @param  obj asarray饹Υ֥
 * @throw  Хåեγݤ˼Ԥ
 * @throw  ˲򵯤Ȥ
 */
template <class datatype>
asarray<datatype>::asarray(const asarray<datatype> &obj)
{
    this->init(obj);
    return;
}

/* destructor */

/**
 * @brief  ǥȥ饯
 *
 */
template <class datatype>
asarray<datatype>::~asarray()
{
    return;
}

template <class datatype>
/**
 * @brief  ֥ȤΥԡ
 *
 *  黻Ҥα¦() ǻꤵ줿Ʊ饹򰷤֥Ȥ򼫿Ȥ
 *  ԡޤ
 * 
 * @param   obj asarray饹Υ֥
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 * @throw   ˲򵯤
 */
asarray<datatype> &asarray<datatype>::operator=(const asarray<datatype> &obj)
{
    this->init(obj);
    return *this;
}

template <class datatype>
/**
 * @brief  [] ǻꤵ줿ͤλȤ֤
 *
 *  ꤵ줿бϢͤλȤ֤ޤ
 * 
 * @param   key ϢΥʸ
 * @return  бϢͤλ
 * @throw   ꤵ줿ʸNULLξ
 * @throw   Хåեγݤ˼Ԥ
 */
inline datatype &asarray<datatype>::operator[]( const char *key )
{
    return this->at(key);
}

template <class datatype>
/**
 * @brief  [] ǻꤵ줿ͤλȤ֤ (ɼ)
 *
 *  ꤵ줿бϢͤλȤ֤ޤ
 * 
 * @param   key ϢΥʸ
 * @return  бϢͤλ
 * @throw   ꤵ줿ʸNULLξ
 * @throw   ¸ߤʤʸ󤬻ꤵ줿
 */
inline const datatype &asarray<datatype>::operator[]( const char *key ) const
{
    return this->at_cs(key);
}

/*
 * public member functions
 */

template <class datatype>
/**
 * @brief  ֥Ȥν
 *
 *  ȤϢõ֥ȤνԤޤ
 * 
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::init()
{
    this->index_rec.init();
    this->key_rec.init();
    this->value_rec.init();
    return *this;
}

template <class datatype>
/**
 * @brief  ֥ȤΥԡ
 * 
 *  ꤵ줿֥ obj Ƥ򼫿Ȥ˥ԡޤ
 * 
 * @param   obj asarray饹Υ֥
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 * @throw   ˲򵯤
 */
asarray<datatype> &asarray<datatype>::init(const asarray<datatype> &obj)
{
    if ( this == &obj ) return *this;
    this->value_rec.init(obj.value_rec);
    this->key_rec.init(obj.key_rec);
    this->index_rec.init(obj.index_rec);
    return *this;
}

template <class datatype>
/**
 * @brief  ¸ͤ٤Ƥꤵ줿ͤǥѥǥ
 *
 *  ȤϢƤǤդͤǥѥǥ󥰤ޤ
 * 
 * @param   one ϢǤѥǥ󥰤
 * @return  Ȥλ
 * @throw   ǤΥХåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::clean(const datatype &one)
{
    this->value_rec.clean(one);
    return *this;
}

template <class datatype>
/**
 * @brief  ꤵ줿Ϣ򼫿Ȥ
 *
 *  srcǻꤵ줿Ϣ򡤼Ȥޤ
 * 
 * @param   src asarray饹Υ֥
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::assign(const asarray<datatype> &src)
{
    if ( this == &src ) return *this;
    this->erase();
    return this->append(src);
}

template <class datatype>
/**
 * @brief  ꤵ줿(key,value) 1ĤΤߤϢȤ
 *
 *  Ȥ򡤻ꤵ줿 1ĤǤϢˤޤ
 * 
 * @param   key Ϣꤹ륭ʸ
 * @param   val Ϣꤹ
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::assign( const char *key, 
					      const datatype &val )
{
    /* ȤƤ key, val Ƥ⤤褦ˤ */
    if ( key != NULL ) {
	size_t org_len = this->length();
	/* ޤǸɲä */
	this->value_rec.append(val,1);
	this->key_rec.append(key,1);
	/* ΤĤõ */
	this->value_rec.erase(0, org_len);
	this->key_rec.erase(0, org_len);
	/* */
	this->index_rec.init();
	this->index_rec.append(this->key_rec.cstr(0), 0);
    }
    else {
	this->erase();
    }
    return *this;
}

template <class datatype>
/**
 * @brief  ꤵ줿ʣʸϢΥȤƼȤ
 *
 *  ꤵ줿ʣʸ򼫿ȤϢΥꤷޤ<br>
 *  Ѱνü NULL ǤʤФʤޤ
 * 
 * @param   key0 ʸ
 * @param   ... ʸβĹΥꥹ (NULLü)
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::assign_keys( const char *key0, ... )
{
    va_list ap;
    va_start(ap,key0);
    try {
	this->vassign_keys(key0,ap);
    }
    catch (...) {
	va_end(ap);
        this->err_throw1(__FUNCTION__,"FATAL","%s","this->vassign_keys() failed");
    }
    va_end(ap);
    return *this;
}

template <class datatype>
/**
 * @brief  ꤵ줿ʣʸϢΥȤƼȤ
 *
 *  ꤵ줿ʣʸ򼫿ȤϢΥꤷޤ<br>
 *  Ѱνü NULL ǤʤФʤޤ
 * 
 * @param   key0 ʸ
 * @param   ap ʸβĹΥꥹ (NULLü)
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::vassign_keys( const char *key0, va_list ap )
{
    tarray_tstring tmp_keys;
    if ( key0 != NULL ) {
	tmp_keys.append(key0,1);
	while ( 1 ) {
	    const char *keyx = va_arg(ap,char *);
	    if ( keyx == NULL ) break;
	    tmp_keys.append(keyx,1);
	}
    }
    this->assign_keys(tmp_keys);
    return *this;
}

template <class datatype>
/**
 * @brief  ꤵ줿ʸϢΥȤƼȤ
 *
 *  ꤵ줿ʣʸ keys 򼫿ȤϢΥꤷޤ<br>
 *  νü NULL ǤʤФʤޤ
 * 
 * @param   keys ʸꤹʸ (NULLü)
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::assign_keys( const char *const *keys )
{
    tarray_tstring tmp_keys;
    if ( this->key_rec.cstrarray() == keys ) return *this;
    tmp_keys.assign(keys);
    this->assign_keys(tmp_keys);
    return *this;
}

template <class datatype>
/**
 * @brief  ꤵ줿ʸϢΥȤƼȤ
 *
 *  ꤵ줿ʣʸ keys 򼫿ȤϢΥꤷޤ<br>
 * 
 * @param   keys ʸꤹʸ
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::assign_keys( const tarray_tstring &keys )
{
    size_t i;

    if ( &(this->key_rec) == &keys ) return *this;

    this->key_rec.erase();
    this->index_rec.init();
    for ( i=0 ; i < keys.length() ; i++ ) {
	/* Ʊ̾硤name, name(1), name(2), ... Τ褦ˤ */
	int ix=0;
	tstring r_key;
	r_key.assign(keys.cstr(i));
	while ( 0 <= this->index_rec.index(r_key.cstr(),0) ) {
	    ix ++;
	    r_key.printf("%s(%d)",keys.cstr(i),ix);
	}
	this->key_rec.append(r_key.cstr(), 1);
	this->index_rec.append(r_key.cstr(), i);
    }
    this->value_rec.resize(this->key_rec.length());

    return *this;
}

template <class datatype>
/**
 * @brief  ʸʬ䤷ơ (¿ǽ)
 *
 *  ʸʬ䤷,ꤷޤʸʬ줿θĿ
 *  ĶϢǤϺޤ. <br>
 *  delims ˤϡ" \t" Τ褦ñʸꥹȤ˲äɽѤ
 *  "[A-Z]" 뤤 "[^A-Z]" Τ褦ʻ꤬ǽǤ<br>
 *  ơ̤ΡʸפǰϤޤ줿ʸʬ䤷ʤ硤
 *  quotations ˡʸפꤷޤ
 * 
 * @param   src_str ʬоݤʸ
 * @param   delims ڤʸޤʸ
 * @param   zero_str Ĺ0ζڤ̤ʸɤΥե饰
 * @param   quotations ơ󡦥֥饱åʸޤʸ
 * @param   escape ʸ
 * @param   rm_escape ʸ뤫ɤΥե饰
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::split_keys( const char *src_str, 
					 const char *delims, 
					 bool zero_str, const char *quotations,
					 int escape, bool rm_escape )
{
    tarray_tstring tmp_keys;

    tmp_keys.split(src_str,delims,zero_str,quotations,escape,rm_escape);
    this->assign_keys(tmp_keys);

    return *this;
}

template <class datatype>
/**
 * @brief  ʸʬ䤷ơ
 *
 *  ʸʬ䤷,ꤷޤʸʬ줿θĿ
 *  ĶϢǤϺޤ. <br>
 *  delims ˤϡ" \t"Τ褦ñʸꥹȤ˲äɽѤ
 *  "[A-Z]" 뤤 "[^A-Z]" Τ褦ʻ꤬ǽǤ
 * 
 * @param   src_str ʬоݤʸ
 * @param   delims ڤʸޤʸ
 * @param   zero_str Ĺ0ζڤ̤ʸɤΥե饰
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::split_keys( const char *src_str, 
						  const char *delims,
						  bool zero_str )
{
    return this->split_keys(src_str,delims,zero_str,NULL,'\0',false);
}

template <class datatype>
/**
 * @brief  ʸʬ䤷ơ (¿ǽ)
 *
 *  ʸʬ䤷ơꤷޤʸʬ줿θĿ
 *  ĶϢǤϺޤ<br>
 *  delims ˤϡ" \t" Τ褦ñʸꥹȤ˲äɽѤ
 *  "[A-Z]" 뤤 "[^A-Z]" Τ褦ʻ꤬ǽǤ<br>
 *  ơ̤ΡʸפǰϤޤ줿ʸʬ䤷ʤ硤
 *  quotations ˡʸפꤷޤ
 *
 * @param   src_str ʬоȤʸ
 * @param   delims ڤʸޤʸ
 * @param   zero_str Ĺ0ζڤ̤ʸɤΥե饰
 * @param   quotations ơ󡦥֥饱åʸޤʸ
 * @param   escape ʸ
 * @param   rm_escape ʸ뤫ɤΥե饰
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::split_keys( const tstring &src_str, 
					 const char *delims, 
					 bool zero_str, const char *quotations,
					 int escape, bool rm_escape )
{
    tarray_tstring tmp_keys;

    tmp_keys.split(src_str,delims,zero_str,quotations,escape,rm_escape);
    this->assign_keys(tmp_keys);

    return *this;
}

template <class datatype>
/**
 * @brief  ʸʬ䤷ơ
 *
 *  ʸʬ䤷ơꤷޤʸʬ줿θĿ
 *  ĶϢǤϺޤ. <br>
 *  delims ˤϡ" \t"Τ褦ñʸꥹȤ˲äɽѤ
 *  "[A-Z]" 뤤 "[^A-Z]" Τ褦ʻ꤬ǽǤ
 * 
 * @param   src_str ʬоݤʸ
 * @param   delims ڤʸޤʸ
 * @param   zero_str Ĺ0ζڤ̤ʸɤΥե饰
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::split_keys( const tstring &src_str, 
						  const char *delims,
						  bool zero_str )
{
    return this->split_keys(src_str,delims,zero_str,NULL,'\0',false);
}

template <class datatype>
/**
 * @brief  ꤵ줿Ϣ򡤼ȤϢɲ
 *
 *  ȤϢˡꤵ줿ʣǤɲäޤ<br>
 *  ʣ硤¹Ի˷ٹ𤬽Ϥ졤Ԥޤ
 * 
 * @param   src ȤʤǤasarary饹Υ֥
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::append( const asarray<datatype> &src )
{
    size_t i;
    if ( this == &src ) {
	/* ʣΤǼȤɲä뤳ȤϤǤʤ */
	this->err_report1(__FUNCTION__,"WARNING", 
			  "%s", "cannot append self-object");
	goto quit;
    }
    for ( i=0 ; i < src.length() ; i++ ) {
	this->append(src.key(i),src.at_cs(i));
    }
 quit:
    return *this;
}

template <class datatype>
/**
 * @brief  ꤵ줿1Ǥ򡤼ȤϢɲ
 *
 *  ȤϢˡꤵ줿 1ĤǤɲäޤ<br>
 *  ʣ硤¹Ի˷ٹ𤬽Ϥ졤Ԥޤ
 * 
 * @param   key Ϣɲä륭ʸ
 * @param   val Ϣɲä
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::append( const char *key, 
					      const datatype &val )
{
    if ( key != NULL ) {
	if ( this->index_rec.index(key,0) < 0 ) {
	    this->value_rec.append(val,1);
	    this->key_rec.append(key,1);
	    this->index_rec.append(key,this->key_rec.length()-1);
	}
	else {
	    this->err_report1(__FUNCTION__,"WARNING",
			      "cannot append key '%s'",key);
	}
    }
    return *this;
}

template <class datatype>
/**
 * @brief  ꤵ줿Ϣ򡤼ȤϢ
 *
 *  ȤϢΥ key ǰ֤ˡꤵ줿ʣǤ
 *  ޤ<br>
 *  ʣ硤¹Ի˷ٹ𤬽Ϥ졤Ԥޤ
 * 
 * @param   key ֤ˤ뼫ȤϢΥʸ
 * @param   src ȤʤǤasarary饹Υ֥
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::insert( const char *key,
					      const asarray<datatype> &src )
{
    size_t i;
    if ( this == &src ) {
	/* ʣΤǼȤ򥤥󥵡Ȥ뤳ȤϤǤʤ */
	this->err_report1(__FUNCTION__,"WARNING", 
			  "%s", "cannot insert self-object");
	goto quit;
    }
    for ( i=0 ; i < src.length() ; i++ ) {
	this->insert(key,src.key(i),src.at_cs(i));
    }
 quit:
    return *this;
}

template <class datatype>
/**
 * @brief  ꤵ줿1Ǥ򡤼ȤϢ
 *
 *  ȤϢΥ key ǰ֤ˡꤵ줿 1ĤǤ
 *  ޤ<br>
 *  ʣ硤¹Ի˷ٹ𤬽Ϥ졤Ԥޤ
 * 
 * @param   key ֤ˤ뼫ȤϢΥʸ
 * @param   newkey Ϣ륭ʸ
 * @param   newval Ϣ
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::insert( const char *key,
				   const char *newkey, const datatype &newval )
{
    if ( key != NULL && newkey != NULL ) {
	size_t i;
	ssize_t idx = this->index_rec.index(key,0);
	if ( idx < 0 ) {
	    this->err_report1(__FUNCTION__,"WARNING","unknown key '%s'",key);
	    goto quit;
	}
	if ( 0 <= this->index_rec.index(newkey,0) ) {
	    this->err_report1(__FUNCTION__,"WARNING",
			      "cannot insert key '%s'",newkey);
	    goto quit;
	}
	this->value_rec.insert(idx,newval,1);
	this->key_rec.insert(idx,newkey,1);
	this->index_rec.append(newkey,idx);
	for ( i=idx+1 ; i < this->key_rec.length() ; i++ ) {
	    this->index_rec.update(this->key_rec.cstr(i),i-1,i);
	}
    }
 quit:
    return *this;
}

template <class datatype>
/**
 * @brief  Ǥκ
 *
 *  ȤϢƤǤõޤ
 * 
 * @return  Ȥλ
 * @hrow    Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::erase()
{
    this->index_rec.init();
    this->key_rec.erase();
    this->value_rec.erase();
    return *this;
}

template <class datatype>
/**
 * @brief  Ǥκ
 *
 *  ȤϢ key ǻꤵ줿бǤ num_elements Ĥ
 *  Ǥõޤ<br>
 *  õ줿ʬĹûʤޤ
 * 
 * @param   key ʸ
 * @param   num_elements ǤθĿ (ά 1)
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::erase( const char *key, 
					     size_t num_elements )
{
    size_t i, maxels;
    ssize_t idx;
    if ( key == NULL ) goto quit;
    idx = this->index_rec.index(key,0);
    if ( idx < 0 ) {
	this->err_report1(__FUNCTION__,"WARNING","unknown key '%s'",key);
	goto quit;
    }
    maxels = this->key_rec.length() - idx;
    if ( maxels < num_elements ) num_elements = maxels;

    for ( i=idx ; i < idx + num_elements ; i++ ) {
	this->index_rec.erase(this->key_rec.cstr(i),i);
    }
    for ( ; i < this->key_rec.length() ; i++ ) {
	this->index_rec.update(this->key_rec.cstr(i),i,i-num_elements);
    }

    this->key_rec.erase(idx,num_elements);
    this->value_rec.erase(idx,num_elements);

 quit:
    return *this;
}

template <class datatype>
/**
 * @brief  2ĤΥ֥ȴ֤ǤƤΥå
 *
 *  ֥sobj ƤȼȤƤȤؤޤ
 * 
 * @param   sobj Ƥؤasarray֥
 * @return  Ȥλ
 */
asarray<datatype> &asarray<datatype>::swap( asarray<datatype> &sobj )
{
    if ( this == &sobj ) return *this;
    this->key_rec.swap( sobj.key_rec );
    this->value_rec.swap( sobj.value_rec );
    this->index_rec.swap( sobj.index_rec );
    return *this;
}

template <class datatype>
/**
 * @brief  ʸѹ
 *
 *  ȤϢΥʸ org_key  new_key ǻꤵ줿ʸѹ
 *  ޤ<br>
 *  org_key ¸ߤʤ䡤new_key ʣ륭ʸäϷٹ
 *  Ϥޤ
 * 
 * @param   org_key Υʸ
 * @param   new_key ѹΥʸ
 * @return  Ȥλ
 * @throw   Хåեγݤ˼Ԥ
 */
asarray<datatype> &asarray<datatype>::rename_a_key( const char *org_key, 
						    const char *new_key )
{
    ssize_t idx;
    idx = this->index_rec.index(org_key,0);
    if ( idx < 0 ) {
	this->err_report1(__FUNCTION__,"WARNING","unknown key '%s'",org_key);
	goto quit;
    }
    if ( new_key == NULL || this->key_rec.at(idx).compare(new_key) == 0 )
	goto quit;
    if ( 0 <= this->index_rec.index(new_key,0) ) {
	this->err_report1(__FUNCTION__,"WARNING",
			  "already used: key '%s'. Not renamed.",new_key);
	goto quit;
    }
    this->index_rec.erase(org_key, idx);
    this->index_rec.append(new_key, idx);
    this->key_rec.at(idx).assign(new_key);
 quit:
    return *this;
}

template <class datatype>
/**
 * @brief  ꤵ줿ʸбǤλȤ֤
 * 
 * @param   key ϢΥʸ
 * @return  ꤵ줿˳ͤλ
 * @throw   ꤵ줿ʸNULLξ
 * @throw   Хåեγݤ˼Ԥ
 */
inline datatype &asarray<datatype>::at( const char *key )
{
    ssize_t idx;
    if ( key == NULL ) {
        this->err_throw1(__FUNCTION__,"ERROR","NULL key...?","");
    }
    idx = this->index_rec.index(key,0);
    if ( idx < 0 ) {
	this->append(key,this->value_rec.initial_one());
	idx = this->index_rec.index(key,0);
    }
    if ( idx < 0 ) {
        this->err_throw1(__FUNCTION__,"ERROR","not found key: %s",key);
    }
    return this->value_rec.at(idx);
}

template <class datatype>
/**
 * @brief  ʸбǤλȤ֤ (printf()εˡ)
 *
 *  ΥʸбϢͤλȤ֤ޤ<br>
 *  ΥдؿǤϻꤷʸ printf() ؿƱͤΥեޥå
 *  ȲѰǥåȤǤޤ.
 * 
 * @param   fmt ʸΤΥեޥåȻ
 * @param   ... fmtбĹγǥǡ
 * @return  ꤵ줿˳ͤλ
 * @throw   ꤵ줿ʸNULLξ
 * @throw   Хåեγݤ˼Ԥ
 */
datatype &asarray<datatype>::atf( const char *fmt, ... )
{
    datatype *ret_p = NULL;
    va_list ap;
    va_start(ap,fmt);
    try {
	datatype &ret = this->vatf(fmt,ap);
	va_end(ap);
	ret_p = &ret;
    }
    catch (...) {
	va_end(ap);
        this->err_throw1(__FUNCTION__,"FATAL","this->vatf() failed","");
    }
    return *(ret_p);
}

template <class datatype>
/**
 * @brief  ʸбǤλȤ֤ (printf()εˡ)
 *
 *  ΥʸбϢͤλȤ֤ޤ.<br>
 *  ΥдؿǤϻꤷʸ printf() ؿƱͤΥեޥå
 *  ȲѰǥåȤǤޤ.
 * 
 * @param   fmt ʸΤΥեޥåȻ
 * @param   ap fmtбĹΥꥹ
 * @return  ꤵ줿˳ͤλ
 * @throw   ꤵ줿ʸNULLξ
 * @throw   Хåեγݤ˼Ԥ
 */
datatype &asarray<datatype>::vatf( const char *fmt, va_list ap )
{
    tstring buf;

    if ( fmt == NULL ) return this->at(fmt);

    buf.vprintf(fmt,ap);
    return this->at(buf.cstr());
}

#ifdef SLI__OVERLOAD_CONST_AT
template <class datatype>
/**
 * @brief  ʸбǤλȤ֤ (ɼ)
 *
 * @param   key ϢΥʸ
 * @return  ꤵ줿˳ͤλ
 * @throw   ꤵ줿ʸNULLξ
 */
inline const datatype &asarray<datatype>::at( const char *key ) const
{
    return this->at_cs(key);
}

template <class datatype>
/**
 * @brief  ʸбǤλȤ֤ (ɼѡprintf()εˡ)
 *
 *  Υʸбͤλ(ɤ߼)֤ޤ<br>
 *  Υдؿϻꤷʸ printf() ؿƱͤΥեޥåȤ
 *  ѰǥåȤǤޤ
 * 
 * @param   fmt ʸΤΥեޥåȻ
 * @param   ... fmtбĹγǥǡ
 * @return  ꤵ줿˳ͤλ
 * @throw   ꤵ줿ʸNULLξ
 * @throw   Хåեγݤ˼Ԥ
 */
const datatype &asarray<datatype>::atf( const char *fmt, ... ) const
{
    const datatype *ret_p = NULL;
    va_list ap;
    va_start(ap,fmt);
    try {
	const datatype &ret = this->vatf_cs(fmt,ap);
	va_end(ap);
	ret_p = &ret;
    }
    catch (...) {
	va_end(ap);
        this->err_throw1(__FUNCTION__,"FATAL","this->vatf_cs() failed","");
    }
    return *(ret_p);
}
template <class datatype>
/**
 * @brief  ʸбǤλȤ֤ (ɼѡprintf()εˡ)
 *
 *  Υʸбͤλ(ɤ߼)֤ޤ<br>
 *  Υдؿϻꤷʸ printf() ؿƱͤΥեޥåȤ
 *  ѰǥåȤǤޤ
 * 
 * @param   fmt ʸΤΥեޥåȻ
 * @param   ap fmtбĹΥꥹ
 * @return  ꤵ줿˳ͤλ
 * @throw   ꤵ줿ʸNULLξ
 * @throw   Хåեγݤ˼Ԥ
 */
const datatype &asarray<datatype>::vatf( const char *fmt, va_list ap ) const
{
    return this->vatf_cs(fmt, ap);
}
#endif

template <class datatype>
/**
 * @brief  ʸбǤλȤ֤ (ɼ)
 * 
 * @param   key ϢΥʸ
 * @return  ꤵ줿˳ͤλ
 * @throw   ¸ߤʤʸ󤬻ꤵ줿
 * @throw   ꤵ줿ʸNULLξ
 */
inline const datatype &asarray<datatype>::at_cs( const char *key ) const
{
    ssize_t idx;
    if ( key == NULL ) {
        this->err_throw1(__FUNCTION__,"ERROR","NULL key...?","");
    }
    idx = this->index_rec.index(key,0);
    if ( idx < 0 ) {
        this->err_throw1(__FUNCTION__,"ERROR","not found key '%s'",key);
    }
    return this->value_rec.at_cs(idx);
}

template <class datatype>
/**
 * @brief  ʸбǤλȤ֤ (ɼѡprintf()εˡ)
 *
 *  Υʸбͤλ(ɤ߼)֤ޤ<br>
 *  Υдؿϻꤷʸ printf() ؿƱͤΥեޥåȤ
 *  ѰǥåȤǤޤ
 * 
 * @param   fmt ʸΤΥեޥåȻ
 * @param   ... fmtбĹγǥǡ
 * @return  ꤵ줿˳ͤλ
 * @throw   ¸ߤʤʸ󤬻ꤵ줿
 * @throw   ꤵ줿ʸNULLξ
 * @throw   Хåեγݤ˼Ԥ
 */
const datatype &asarray<datatype>::atf_cs( const char *fmt, ... ) const
{
    const datatype *ret_p = NULL;
    va_list ap;
    va_start(ap,fmt);
    try {
	const datatype &ret = this->vatf_cs(fmt,ap);
	va_end(ap);
	ret_p = &ret;
    }
    catch (...) {
	va_end(ap);
        this->err_throw1(__FUNCTION__,"FATAL","this->vatf_cs() failed","");
    }
    return *(ret_p);
}

template <class datatype>
/**
 * @brief  ʸбǤλȤ֤ (ɼѡprintf()εˡ)
 *
 *  Υʸбͤλ(ɤ߼)֤ޤ<br>
 *  Υдؿϻꤷʸ printf() ؿƱͤΥեޥåȤ
 *  ѰǥåȤǤޤ
 * 
 * @param   fmt ʸΤΥեޥåȻ
 * @param   ap fmtбĹΥꥹ
 * @return  ꤵ줿˳ͤλ
 * @throw   ¸ߤʤʸ󤬻ꤵ줿
 * @throw   ꤵ줿ʸNULLξ
 * @throw   Хåեγݤ˼Ԥ
 */
const datatype &asarray<datatype>::vatf_cs( const char *fmt, va_list ap ) const
{
    tstring buf;

    if ( fmt == NULL ) return this->at_cs(fmt);

    buf.vprintf(fmt,ap);
    return this->at_cs(buf.cstr());
}

template <class datatype>
/**
 * @brief  ʸ󥪥֥ȤλȤ (ɼ)
 *
 *  ֥ǴƤ륭ʸ󥪥֥Ȥλ(ɤ߼
 *  )֤ޤ
 * 
 * @return  ʸ󥪥֥
 */
inline const tarray_tstring &asarray<datatype>::keys() const
{
    return this->key_rec;
}

template <class datatype>
/**
 * @brief  ͤ󥪥֥ȤλȤ (ɼ)
 *
 *  ֥ǴƤͤ󥪥֥Ȥλ(ɤ߼)
 *  ֤ޤ
 * 
 * @return  ͤ󥪥֥Ȥλ
 */
inline const tarray<datatype> &asarray<datatype>::values() const
{
    return this->value_rec;
}

template <class datatype>
/**
 * @brief  ϢĹ(Ŀ)
 * 
 * @return  ȤϢθĿ
 */
inline size_t asarray<datatype>::length() const
{
    return this->key_rec.length();
}

template <class datatype>
/**
 * @brief  ϢĹ(Ŀ)
 * 
 * @return  ȤϢθĿ
 * @note    asarray<datatype>::length() Ȥΰ㤤Ϥޤ
 */
inline size_t asarray<datatype>::size() const
{
    return this->key_rec.size();
}

template <class datatype>
/**
 * @brief  ꤵ줿ʸбֹ֤
 * 
 * @param   key ʸ
 * @return  : ꤵ줿ʸ󤬸Ĥä硥<br>
 *          : Ĥʤä硥
 */
ssize_t asarray<datatype>::index( const char *key ) const
{
    if ( key == NULL ) return -1;
    return this->index_rec.index(key,0);
}

template <class datatype>
/**
 * @brief  ꤵ줿ʸбֹ֤ (printf()εˡ)
 *
 *  ʸбֹ֤ޤ<br>
 *  ΥдؿǤ printf() ؿƱͤΥեޥåȤȲѰǡ
 *  ʸ򥻥åȤǤޤ
 * 
 * @param   fmt ʸΤΥեޥåȻ
 * @param   ... fmtбĹγǥǡ
 * @return  : ꤵ줿ʸ󤬸Ĥä硥<br>
 *          : Ĥʤä硥
 * @throw   Хåեγݤ˼Ԥ
 */
ssize_t asarray<datatype>::indexf( const char *fmt, ... ) const
{
    ssize_t ret = -1;
    va_list ap;
    va_start(ap,fmt);
    try {
	ret = this->vindexf(fmt,ap);
    }
    catch (...) {
	va_end(ap);
        this->err_throw1(__FUNCTION__,"FATAL","this->vindexf() failed","");
    }
    va_end(ap);
    return ret;
}

template <class datatype>
/**
 * @brief  ꤵ줿ʸбֹ֤ (printf()εˡ)
 *
 *  ʸбֹ֤ޤ<br>
 *  ΥдؿǤ printf() ؿƱͤΥեޥåȤȲѰǡ
 *  ʸ򥻥åȤǤޤ
 * 
 * @param   fmt ʸΤΥեޥåȻ
 * @param   ap fmtбĹΥꥹ
 * @return  : ꤵ줿ʸ󤬸Ĥä硥<br>
 *          : Ĥʤä硥
 * @throw   Хåեγݤ˼Ԥ
 */
ssize_t asarray<datatype>::vindexf( const char *fmt, va_list ap ) const
{
    tstring buf;

    if ( fmt == NULL ) return this->index(fmt);

    buf.vprintf(fmt,ap);
    return this->index(buf.cstr());
}

template <class datatype>
/**
 * @brief  ꤵ줿ֹ˳륭ʸ֤
 * 
 *  ꤵ줿ֹ˳륭ʸΥɥ쥹֤ޤ
 * 
 * @param   index ֹ
 * @return  ʸХåեΥɥ쥹<br>
 *          Ĺʾֹ椬ꤵ줿NULL
 */
const char *asarray<datatype>::key( size_t index ) const
{
    if ( index < 0 || this->length() <= index ) return NULL;
    else return this->key_rec.cstr(index);
}

template <class datatype>
/**
 * @brief  ꤵ줿ֹ˳ͤλȤ֤
 * 
 * @param   index ֹ
 * @return  ꤵ줿ֹ˳ͤλ
 * @throw   ꤵ줿ֹ椬ʾ
 */
inline datatype &asarray<datatype>::at( size_t index )
{
    static tstring buf(32);
    if ( index < 0 || this->length() <= index ) {
	buf.printf("%d",(int)index);
        this->err_throw1(__FUNCTION__,"ERROR","invalid index: %s",buf.cstr());
    }
    return this->value_rec.at(index);
}


template <class datatype>
/**
 * @brief  ꤵ줿ֹ˳ͤλ(tstring)֤ (ɼ)
 * 
 * @param   index ֹ
 * @return  ꤵ줿ֹ˳ͤλ
 * @throw   ꤵ줿ֹ椬ʾ
 */
inline const datatype &asarray<datatype>::at_cs( size_t index ) const
{
    static tstring buf(32);
    if ( index < 0 || this->length() <= index ) {
	buf.printf("%d",(int)index);
        this->err_throw1(__FUNCTION__,"ERROR","invalid index: %s",buf.cstr());
    }
    return this->value_rec.at_cs(index);
}

/* new datatype ˡ󥹥ȥ饯˰Ϳ˻Ȥ */

/**
 * @brief  datatype饹Υ֥ȤؿϿ
 *
 *  asarray 饹ǡnew datatypeפˡ󥹥ȥ饯˰
 *  Ϳ˻Ȥޤ.
 *
 * @param     func 桼ؿΥɥ쥹
 * @param     user_ptr 桼ؿκǸͿ桼Υݥ
 * @throw     Хåեγݤ˼Ԥ
 */
template <class datatype>
void asarray<datatype>::register_creator( datatype *(*func)(void *), 
					  void *user_ptr )
{
    this->value_rec.register_creator(func, user_ptr);
    return;
}

/**
 * @brief  ƥݥꥪ֥ȤΤshallow copy°Ϳ (̤)
 * @deprecated  ̤
 */
/* : Ȥ֤Ȥreturn ʸǤȥ󥹥 */
/*       2󥳥ԡȤʻˤʤΤ void ˤʤäƤ */
template <class datatype>
void asarray<datatype>::set_scopy_flag()
{
    return;
}


/* 󡤤ȤϤǤʤ餷orz */
#if 0
template <class datatype>
asarray<datatype> &asarray<datatype>::assign( const char *key0, 
					      const datatype &val0,
					      const char *key1, ... )
{
    va_list ap;
    va_start(ap,key1);
    try {
	this->vassign(key0,val0,key1,ap);
    }
    catch (...) {
	va_end(ap);
        this->err_throw1(__FUNCTION__,"FATAL","this->vassign() failed","");
    }
    va_end(ap);
    return *this;
}
template <class datatype>
asarray<datatype> &asarray<datatype>::vassign( const char *key0, 
					       const datatype &val0,
					       const char *key1, va_list ap )
{
    this->assign(key0,val0);
    if ( key0 != NULL && key1 != NULL ) {
	const datatype &val1 = va_arg(ap,const datatype &);
	this->append(key1,val1);
	while ( 1 ) {
	    const char *keyx = va_arg(ap,const char *);
	    if ( keyx == NULL ) break;
	    const datatype &valx = va_arg(ap,const datatype &);
	    this->append(keyx,valx);
	}
    }
    return *this;
}
template <class datatype>
asarray<datatype> &asarray<datatype>::append( const char *key0, 
					      const datatype &val0,
					      const char *key1, ... )
{
    va_list ap;
    va_start(ap,key1);
    try {
	this->vappend(key0,val0,key1,ap);
    }
    catch (...) {
	va_end(ap);
        this->err_throw1(__FUNCTION__,"FATAL","this->vappend() failed","");
    }
    va_end(ap);
    return *this;
}
template <class datatype>
asarray<datatype> &asarray<datatype>::vappend( const char *key0, 
					       const datatype &val0,
					       const char *key1, va_list ap )
{
    this->append(key0,val0);
    if ( key0 != NULL && key1 != NULL ) {
	const datatype &val1 = va_arg(ap,const datatype &);
	this->append(key1,val1);
	while ( 1 ) {
	    const char *keyx = va_arg(ap,char *);
	    if ( keyx == NULL ) break;
	    const datatype &valx = va_arg(ap,const datatype &);
	    this->append(keyx,valx);
	}
    }
    return *this;
}
template <class datatype>
asarray<datatype> &asarray<datatype>::insert( const char *key,
					const char *key0, const datatype &val0,
					const char *key1, ... )
{
    va_list ap;
    va_start(ap,key1);
    try {
	this->vinsert(key,key0,val0,key1,ap);
    }
    catch (...) {
	va_end(ap);
        this->err_throw1(__FUNCTION__,"FATAL","this->vinsert() failed","");
    }
    va_end(ap);
    return *this;
}
template <class datatype>
asarray<datatype> &asarray<datatype>::vinsert( const char *key,
					const char *key0, const datatype &val0,
					const char *key1, va_list ap )
{
    if ( key == NULL ) goto quit;
    if ( this->index_rec.index(key,0) < 0 ) {
	this->err_report1(__FUNCTION__,"WARNING","unknown key '%s'",key);
	goto quit;
    }
    this->insert(key,key0,val0);
    if ( key0 != NULL && key1 != NULL ) {
	const datatype &val1 = va_arg(ap,const datatype &);
	this->insert(key,key1,val1);
	while ( 1 ) {
	    const char *keyx = va_arg(ap,char *);
	    if ( keyx == NULL ) break;
	    const datatype &valx = va_arg(ap,const datatype &);
	    this->insert(key,keyx,valx);
	}
    }
 quit:
    return *this;
}
#endif


/* private member functions */

/**
 * @brief  顼㳰ȯ err_rec  throw 
 *
 *  asarray饹δؿǥ顼㳰ȯ˸ƤӽФޤ
 *  ɸ२顼Ϥ˥顼ɽerr_rec  throwޤ
 *
 * @param	func0 ƤӽФؿ̾
 * @param	level0 顼<BR>
 *			㳰ȯ:"FATAL"<BR>
 *			顼:"ERROR"
 * @param	message0 顼å
 * @param	arg 顼åǥǡ
 * @throw	err_rec:顼Ĺ¤.
 * @note	ΥдؿprivateǤ
 */
template <class datatype>
void asarray<datatype>::err_throw1( const char *func0, const char *level0, 
				  const char *message0, const char *arg ) const
{
    static stdstreamio sio;
    static tstring mes(128);
    const char *classname = "asarray";
    sio.eprintf("%s::%s(): [%s] ",classname,func0,level0);
    mes.printf(message0, arg);
    sio.eprintf("%s\n", mes.cstr());
    {
	size_t i, len;
	const char *src;
	char *dst;
	sli::err_rec st;
	/* */
	src = classname;  dst = st.class_name;   len = 64;
	for ( i=0 ; i < len && src[i] != '\0' ; i++ ) dst[i] = src[i];
	dst[i] = '\0';
	/* */
	src = func0;      dst = st.func_name;    len = 64;
	for ( i=0 ; i < len && src[i] != '\0' ; i++ ) dst[i] = src[i];
	dst[i] = '\0';
	/* */
	src = level0;     dst = st.level;        len = 64;
	for ( i=0 ; i < len && src[i] != '\0' ; i++ ) dst[i] = src[i];
	dst[i] = '\0';
	/* */
	src = mes.cstr(); dst = st.message;      len = 128;
	for ( i=0 ; i < len && src[i] != '\0' ; i++ ) dst[i] = src[i];
	dst[i] = '\0';
	/* */
	throw st;
    }
}


/**
 * @brief  ٹϤ뤿Υдؿ
 *
 *  asarray饹ΥдؿǷٹȯȤ˸ƤӽФޤ
 *  ɸ२顼Ϥ˥顼åϤޤ
 *
 * @param	func0 ƤӽФдؿ̾
 * @param	level0 ٹ:"WARNING"
 * @param	message0 顼åν
 * @param	arg	顼åǥǡ
 * @note	ΥдؿprivateǤ
 */
template <class datatype>
void asarray<datatype>::err_report1( const char *func0, const char *level0, 
				  const char *message0, const char *arg ) const
{
    static stdstreamio sio;
    const char *classname = "asarray";
    sio.eprintf("%s::%s(): [%s] ",classname,func0,level0);
    sio.eprintf(message0, arg);
    sio.eprintf("\n");
}

}	/* namespace sli */

#endif	/* _SLI__ASARRAY_H */
