/* -*- Mode: C++ ; Coding: euc-japan -*- */
/* Time-stamp: <2013-04-27 19:48:24 cyamauch> */

#ifndef _SLI__CTINDEX_H
#define _SLI__CTINDEX_H 1

/**
 * @file   ctindex.h
 * @brief  ʸֹȤδط륯饹 ctindex Υإåե
 */

#include <stddef.h>
#include <sys/types.h>

#include "slierr.h"

/*
 * Character Tree Index
 *
 * ʸ  index(size_t) κCharacter Tree Ȥäƺ롥
 *
 */

namespace sli
{

/*
 * sli::ctindex class manages the relationship between the key string and the
 * index.  Once the relationship between a string (key) and a number (index) is
 * registered, the index can be retrieved quickly by using key string.  Its
 * behavior is similar to that of hash, but no collision occurs because this 
 * class manages the relationship between the key string and the index with a 
 * dictionary.
 */

/**
 * @class  sli::ctindex
 * @brief  ʸȥǥå(ʤ)Ȥδط륯饹
 *
 *   ctindex 饹ϡʸȥǥå(ʤ)Ȥδط
 *   ڡ׹¤ǵϿ®ʸ(ʸ󤫤ֹؤѴ)ǽǤ
 *   ƱΥʸФʣΥǥåϿǤޤ<br>
 *   ˤ֤ϡϿ줿ʸοǤϤʤơΥʸ
 *   Ĺ㤷ޤ(ʸĹۤ®)ڡ׹¤
 *   ǤΤǡˤϤ֤ޤ
 *   
 * @author  Chisato YAMAUCHI
 * @date  2013-04-01 00:00:00
 */

  class ctindex
  {

  public:
    /* constructor and destructor */
    ctindex();
    ctindex(const ctindex &obj);
    virtual ~ctindex();

    /*
     * operator
     */
    virtual ctindex &operator=(const ctindex &obj);

    /*
     * member functions
     */
    /* complete initialization of objects */
    virtual ctindex &init();
    virtual ctindex &init(const ctindex &obj);

    /* append an index */
    virtual int append( const char *key, size_t index );

    /* update an index associate by a key */
    virtual int update( const char *key, 
			size_t current_index, size_t new_index );

    /* erase an index */
    virtual int erase( const char *key, size_t index );

    /* returns an index */
    virtual ssize_t index( const char *key, int index_of_index = 0 ) const;

    /* swap self and another */
    virtual ctindex &swap( ctindex &sobj );

    /* shallow copy Ĥ˻Ȥ (̤) */
    virtual void set_scopy_flag();

  private:
    ctindex *configure_ctable( int ch );
    int realloc_index_rec( size_t len_elements );
    void free_index_rec();
    int realloc_ctable_rec( size_t len_elements );
    void free_ctable_rec();
    bool request_shallow_copy( ctindex *from_obj ) const;

  private:
    size_t *_index_rec;		/* Ū index */
    size_t index_alloc_blen_rec;
    int index_rec_size;		/* index_rec ο */
    int ctable_begin_ch;	/* ctable γϥ饯('A'Ȥ) */
    int ctable_size;		/* ctable Υ('A'  'E' ʤ 5) */
    ctindex **_ctable_rec;
    size_t ctable_alloc_blen_rec;

    /* shallow copy ΤΥե饰= Ǥϥԡʤ */
    /* ԡ src ¦Υե饰 */
    bool shallow_copy_ok;		/* set_scopy_flag() ǥåȤ */
    ctindex *shallow_copy_dest_obj;	/* ԡ: ͭʤ non-NULL */

    /* ԡ dest ¦Υե饰(ԡ񤭹ޤ) */
    ctindex *shallow_copy_src_obj;	/* ԡ: Ƥ non-NULL */

    /* __shallow_init(), __deep_init() ˹ԤʤΥե饰 */
    bool __copying;

  };
}

#endif  /* _SLI__CTINDEX_H */
