/* -*- Mode: C++ ; Coding: euc-japan -*- */
/* Time-stamp: <2013-06-05 11:24:31 cyamauch> */

#ifndef _SLI__DIGESTSTREAMIO_H
#define _SLI__DIGESTSTREAMIO_H 1

/**
 * @file   digeststreamio.h
 * @brief  ̾̥ȥ꡼бǽʥ饹 digeststreamio 
 */

#include "cstreamio.h"
#include "tstring.h"

namespace sli
{

  /*
   * switch of multi-thread compressor support
   */
  /*
   * ޥåб compressor/decompressor 򤳤Ǥ롥
   * **_auto() ξ硤ˡĶѿ on/off Ǥ롥ξ硤
   * SLI_PIGZSLI_LBZIP2  0 ꤹȡѤʤˤʤ롥
   * pigzlbzip2 ؤΰϡδĶѿɲäǤ롥
   *   SLI_PIGZ_R_ARGS   ... pigz ɤ߹
   *   SLI_PIGZ_W_ARGS   ... pigz 񤭹
   *   SLI_LBZIP2_R_ARGS ... lbzip2 ɤ߹
   *   SLI_LBZIP2_W_ARGS ... lbzip2 񤭹
   * 㤨Сpigz ȤäŸϡΤ褦ʥޥɤ¹Ԥ롥
   *   pigz -dc $SLI_PIGZ_R_ARGS filename.gz
   *
   * : 餫θ(㤨 mode  755 ˤʤäƤʤ)ǥץब
   * ưǤʤä open() Ǥϥ顼ȽǤʤʤɡpigz, lbzip2
   * ѤξϾܺ٤ʥ顼åǤʤȤ¤롥pigz, lbzip2
   * ֤ͤ䡤֤ͤϡclose() дؿ֤ͤȤ֤
   * ǡå뤳ȡ
   */
  extern void pigz_auto();		/* use pigz for local files */
  extern void pigz_disabled();		/* do not use pigz */
  extern void lbzip2_auto();		/* use lbzip2 for local files */
  extern void lbzip2_disabled();	/* do not use lbzip2 */

/*
 * digeststreamio class -- a transparent stream I/O of local files and files
 *                         on HTTP and FTP servers with gzip and bzip2 
 *                         compression / decompression support.
 *                         (concrete derived class of cstreamio)
 *
 * linker options : -lz -lbz2 (Linux)
 *                  -lz -lbz2 -lsocket (Solaris)
 *
 */

/**
 * @class  sli::digeststreamio
 * @brief  ̾ե롤HTTPФΰ̥ȥ꡼бǽʥ饹
 *
 *   digeststreamio饹ϡopen()дؿ path Ǽ줿 URL ޤ
 *   եɬפ˱ gzip ޤ bzip2 ̡ĹԤʤʤ饹
 *   ꡼ϤԤʤޤopenp()дؿȤСѥϤǽ
 *   Ǥ<br>
 *   ե pigzlbzip2 Ǥι®ʰ̡ĹбƤޤ
 *
 * @attention  LIBC  stdio.h ʻѤ硤digeststreamio.h 
 *             stdio.h  include Ƥ
 * @attention  ݴ쥯饹Ǥ cstreamio ѾƤޤ<br>
 *             Ū API ˤĤƤ cstreamio 饹 public дؿȤ
 *             ƤޤΤǡ
 * @author  Chisato YAMAUCHI
 * @date  2013-04-01 00:00:00
 */

  class digeststreamio : public cstreamio
  {

  public:
    /* constructor & destructor */
    digeststreamio();
    ~digeststreamio();

    /*
     * overridden member functions:  See cstreamio.h for details.
     */
    /* open() and openp() member functions automatically setup internal     */
    /* operation mode using prefix/suffix of path such as "http://", ".gz", */
    /* etc.                                                                 */
    int open( const char *mode );
    int open( const char *mode, const char *path );
    int open( const char *mode, int fd );
    int close();
    int flush();
    int eof();
    int error();
    digeststreamio &reseterr();

    /*
     * added member functions
     */
    /* these APIs accept Perl-like arguments                 */
    /* e.g., openp("./my_program |"), openp("> outfile.txt") */
    virtual int openp( const char *path );
    virtual int openpf( const char *path_fmt, ... );
    virtual int vopenpf( const char *path_fmt, va_list ap );

    /* this returns true for stream opened as write mode */
    virtual bool is_write_mode() const;

    /* returns length of stream when accessing FTP or HTTP server */
    virtual long long content_length() const;

    /* set a string when sending a value for `User-Agent:' header (for HTTP) */
    virtual tstring &user_agent();

    /* set username and password for non-anonymous login (for FTP) */
    virtual tstring &username();
    virtual tstring &password();

  protected:
    ssize_t raw_read( void *buf, size_t size );
    ssize_t raw_write( const void *buf, size_t size );
    /* added member functions */
    virtual int open_rw( const char *mode, const char *path );

  private:
    cstreamio *isp;	/* httpstreamio, gzstreamio or bzstreamio */
    bool is_httpstream;
    bool is_ftpstream;
    bool is_write_mode_rec;
    /* */
    tstring user_agent_rec;	/* for httpstreamio */
    tstring username_rec;	/* for ftpstreamio */
    tstring password_rec;	/* for ftpstreamio */

  };

/**
 * @example  examples_sllib/read_text_from_net.cc
 *           httpФ ftpоˤƥȥե1Ԥɤߡ
 *           Ϥ륳
 */

}

#endif	/* _SLI__DIGESTSTREAMIO_H */
