/* -*- Mode: C++ ; Coding: euc-japan -*- */
/* Time-stamp: <2013-05-13 23:31:18 cyamauch> */

/**
 * @file   ftpstreamio.cc
 * @brief  FTPФ³ȥ꡼ϤԤʤ ftpstreamio 饹Υ
 */

#define CLASS_NAME "ftpstreamio"

#include "config.h"

#include "ftpstreamio.h"
#include "inetstreamio.h"
#ifdef HAVE_LIBZ
#include "gzstreamio.h"
#endif
#ifdef HAVE_LIBBZ2
#include "bzstreamio.h"
#endif
#include "tstring.h"

#include <assert.h>

#include "private/err_report.h"


static void get_user_at_host( sli::tstring *ret );

#include "private/c_gethostname.h"

#include "private/c_fdopen.h"
#include "private/c_fclose.h"

#include "private/c_strcpy.h"
#include "private/c_strncmp.h"

#include "private/c_geteuid.h"
#include "private/c_dup.h"
#include "private/c_close.h"

#include "private/suffix2filetype.h"


namespace sli
{

/**
 * @brief  󥹥ȥ饯
 */
ftpstreamio::ftpstreamio()
{
    this->inet_sp = NULL;
    this->rw_sp = NULL;
    this->filter_sp = NULL;
    this->content_length_rec = -1;
    return;
}

/**
 * @brief  ǥȥ饯
 */
ftpstreamio::~ftpstreamio()
{
    this->close();
    return;
}

/**
 * @brief  ftp:// Ϥޤ URL Υץ
 *
 *  ѥꤷơURL򥪡ץ󤷤ޤ<br>
 *  mode  "r%" ޤ "w%" ꤹȡե̾Υեåˤꡤ
 *  ưŪ˿Ĺ̤Ԥʤޤ
 *
 * @param      mode URL򳫤⡼ ("r", "r%", "w" or "w%")
 * @param      path URLΥѥ
 * @return     0<br>
 *             顼ξ
 * @throw      γƤ˼Ԥ 
 */
int ftpstreamio::open( const char *mode, const char *path )
{
    assert( mode != NULL );

    int return_status = -1;

    /* "r%" ꤹḁ̇̄ȥ꡼Ÿб */
    if ( c_strncmp(mode,"r%",2) == 0 || c_strncmp(mode,"rb%",3) == 0 ) {
        return_status = this->open_rw(mode,path,true);
    } else if ( c_strncmp(mode,"w%",2) == 0 || c_strncmp(mode,"wb%",3) == 0 ) {
        return_status = this->open_rw(mode,path,true);
    } else if ( c_strncmp(mode,"r",1) == 0 || c_strncmp(mode,"rb",2) == 0 ) {
        return_status = this->open_rw(mode,path,false);
    } else if ( c_strncmp(mode,"w",1) == 0 || c_strncmp(mode,"wb",2) == 0 ) {
        return_status = this->open_rw(mode,path,false);
    } else {
        return return_status = -1;
    }

    return return_status;
}

/**
 * @brief  FTPɤ֤
 *
 * @param      isp inetstreamio饹Υ󥹥
 * @param      rline tstring饹Υ󥹥
 * @return     400̤3<br>
 *             Ԥ400ʾ
 * @note       Υдؿ private Ǥ
 */
int ftpstreamio::read_response( inetstreamio *isp, tstring &rline )
{
    assert( isp != NULL );

    while ( 1 ) {
	const char *l;
	if ( (l=isp->getline()) == NULL ) {
	    rline.assign("999 Connection broken");
	    return 999;
	}
	rline.assign(l);
	if ( rline.strspn("[0-9]") == 3 ) {
	    if ( rline.cchr(3) == ' ' ) {
		//err_report1("DEBUG","DEBUG","[%s]\n",rline.cstr());
		break;
	    }
	}
    }
    return rline.atoi(0,3);
}

/**
 * @brief  URLΥץ (٥)
 *
 * @param      mode URL򳫤⡼ ("r", "r%", "w" or "w%")
 * @param      path URLΥѥ
 * @param      use_filter_by_suffix ̥ȥ꡼Ÿ̵ͭ
 * @return     0<br>
 *             顼ξ
 * @throw      γƤ˼Ԥ 
 * @note       Υдؿ protected Ǥ
 */
int ftpstreamio::open_rw( const char *mode, const char *path,
			  bool use_filter_by_suffix )
{
    int return_status = -1;
    inetstreamio *isp = NULL;	/* inetstreamio (message) */
    inetstreamio *rwsp = NULL;	/* inetstreamio (stream)  */
    cstreamio *dsp = NULL;	/* gzstreamio or bzstreamio */
    char rw_mode[2] = { 'r', '\0' };

    if ( mode == NULL ) goto quit;
    if ( path == NULL ) goto quit;
    if ( c_strncmp(path,"ftp://",6) != 0 ) goto quit;
    if ( this->inet_sp != NULL ) goto quit;
    if ( this->rw_sp != NULL ) goto quit;

    if ( mode[0] == 'w' ) rw_mode[0] = 'w';

    try {
	bool path_is_gzip = false, path_is_bzip2 = false;
	int ip0,ip1,ip2,ip3,port;
	tstring t_path;
	tstring user_pass;
	ssize_t pos;
	size_t len;
	tstring user;
	tstring pass;
	tstring ans;

	isp = new inetstreamio;
	rwsp = new inetstreamio;

	t_path.assign(path);

	/* ftp://username:password@ftp.com/... Ǥ褦ˤ */
	user_pass.assign(path);
	user_pass.erase(0,6);
	pos = user_pass.regmatch("^[^:][^:]*:[^@]*@",&len);
	if ( pos == 0 ) {
	    t_path.erase(6,len);
	    user_pass.crop(pos,len-1);
	}

	/* FTP 桼̾ */
	if ( pos == 0 ) {
	    user.assign(user_pass,0,user_pass.strchr(':'));
	}
	else if ( 0 < this->username_rec.length() &&
		  this->username_rec.strpbrk("[[:cntrl:]]") < 0 ) {
	    user.assign(this->username_rec.cstr());
	}
	else {
	    user.assign("anonymous");
	}

	/* FTP ѥ */
	if ( pos == 0 ) {
	    pass.assign(user_pass,user_pass.strchr(':')+1);
	}
	else if ( 0 < this->password_rec.length() &&
		  this->password_rec.strpbrk("[[:cntrl:]]") < 0 ) {
	    pass.assign(this->password_rec.cstr());
	}
	else {
	    get_user_at_host(&pass);
	}

	//err_report1(__FUNCTION__,"DEBUG","user : %s",user.cstr());
	//err_report1(__FUNCTION__,"DEBUG","pass : %s",pass.cstr());
	//err_report1(__FUNCTION__,"DEBUG","url  : %s",t_path.cstr());

	suffix2filetype(t_path.cstr(),&path_is_gzip,&path_is_bzip2);

#ifndef HAVE_LIBZ
	path_is_gzip = false;
#endif
#ifndef HAVE_LIBBZ2
	path_is_bzip2 = false;
#endif

	if ( isp->open("r+",t_path.cstr()) < 0 ) {
	    err_report1(__FUNCTION__,"ERROR","cannot open url: %s",t_path.cstr());
	    goto quit;
	}
	
	/* ϤΥå */
	if ( 400 <= this->read_response(isp,ans) ) {
	    err_report1(__FUNCTION__,"ERROR","ftpd: [%s]",ans.strtrim().cstr());
	    goto quit;
	}
	
	/* 桼 */
	isp->printf("USER %s\n",user.cstr());
	isp->flush();
	if ( 400 <= this->read_response(isp,ans) ) {
	    err_report1(__FUNCTION__,"ERROR","ftpd: [%s]",ans.strtrim().cstr());
	    goto quit;
	}

	/* ѥ */
	isp->printf("PASS %s\n",pass.cstr());
	isp->flush();
	if ( 400 <= this->read_response(isp,ans) ) {
	    err_report1(__FUNCTION__,"ERROR","ftpd: [%s]",ans.strtrim().cstr());
	    goto quit;
	}

	/* Хʥ */
	isp->printf("TYPE I\n");
	isp->flush();
	if ( 400 <= this->read_response(isp,ans) ) {
	    err_report1(__FUNCTION__,"ERROR","ftpd: [%s]",ans.strtrim().cstr());
	    goto quit;
	}

	/* ե륵 */
	if ( rw_mode[0] == 'r' ) {
	    long long file_size;
	    isp->printf("SIZE %s\n",isp->path());
	    isp->flush();
	    if ( 400 <= this->read_response(isp,ans) ) {
		err_report1(__FUNCTION__,"ERROR","ftpd: [%s]",
			    ans.strtrim().cstr());
		goto quit;
	    }
	    file_size = ans.atoll(4);
	    if ( 0 <= file_size ) this->content_length_rec = file_size;
	}
	
	/* ѥå֥⡼ */
	isp->printf("PASV\n");
	isp->flush();
	if ( 400 <= this->read_response(isp,ans) ) {
	    err_report1(__FUNCTION__,"ERROR","ftpd: [%s]",ans.strtrim().cstr());
	    goto quit;
	}
	pos = ans.regmatch("[0-9][0-9]*,[0-9][0-9]*,"
			   "[0-9][0-9]*,[0-9][0-9]*,"
			   "[0-9][0-9]*,[0-9][0-9]*", NULL);
	if ( pos < 0 ) {
	    err_report1(__FUNCTION__,"ERROR","Invalid server response : [%s]",
			ans.strtrim("\r\n").cstr());
	    goto quit;
	}
	ip0 = ans.atoi(pos);
	pos += ans.strspn(pos,"[0-9]") + 1;
	ip1 = ans.atoi(pos);
	pos += ans.strspn(pos,"[0-9]") + 1;
	ip2 = ans.atoi(pos);
	pos += ans.strspn(pos,"[0-9]") + 1;
	ip3 = ans.atoi(pos);
	pos += ans.strspn(pos,"[0-9]") + 1;
	port = ans.atoi(pos) * 256;
	pos += ans.strspn(pos,"[0-9]") + 1;
	port += ans.atoi(pos);

	if ( rwsp->openf(rw_mode,"ftp://%d.%d.%d.%d:%d/",
			 ip0,ip1,ip2,ip3,port) < 0 ) {
	    err_report1(__FUNCTION__,"ERROR","cannot open port: %d",port);
	    goto quit;
	}

	/*  or ޥ */
	if ( rw_mode[0] == 'r' ) {
	    isp->printf("RETR %s\n",isp->path());
	}
	else {
	    isp->printf("STOR %s\n",isp->path());
	}
	isp->flush();

	/* xxx Opening ... */
	if ( 400 <= this->read_response(isp,ans) ) {
	    err_report1(__FUNCTION__,"ERROR","ftpd: [%s]",ans.strtrim().cstr());
	    goto quit;
	}

	if ( use_filter_by_suffix != false &&
	     ( path_is_gzip != false || path_is_bzip2 != false ) ) {
	    if ( path_is_gzip != false ) {
#ifdef HAVE_LIBZ
		dsp = new gzstreamio;
#endif
	    }
	    else if ( path_is_bzip2 != false ) {
#ifdef HAVE_LIBBZ2
		dsp = new bzstreamio;
#endif
	    }
	    if ( dsp != NULL ) {
	      if ( dsp->open(rw_mode,*rwsp) < 0 ) {
		err_report(__FUNCTION__,"ERROR","dsp->open() failed");
		goto quit;
	      }
	    }
	    else {
	      /* never */
	    }
	}

    }
    catch ( ... ) {
	if ( dsp != NULL ) delete dsp;
	if ( rwsp != NULL ) delete rwsp;
	if ( isp != NULL ) delete isp;
	err_throw(__FUNCTION__,"FATAL","Memory allocation failed");
    }

    return_status = 0;
 quit:
    if ( return_status < 0 ) {
	if ( dsp != NULL ) {
	    dsp->close();
	    delete dsp;
	}
	if ( rwsp != NULL ) {
	    rwsp->close();
	    delete rwsp;
	}
	if ( isp != NULL ) {
	    isp->close();
	    delete isp;
	}
	this->content_length_rec = -1;
    }
    else {
	if ( isp != NULL ) this->inet_sp = isp;
	if ( rwsp != NULL ) this->rw_sp = rwsp;
	if ( dsp != NULL ) this->filter_sp = dsp;
    }
    return return_status;
}

/**
 * @brief  ȥ꡼ΥХĹ
 *
 *  FTP Ф SIZE ޥɤǼͤ򤽤Τޤ֤ޤ
 *
 * @return     : SIZE ޥɤǼ<br>
 *             : SIZE ޥɤ˼Ԥ
 */
long long ftpstreamio::content_length() const
{
    return this->content_length_rec;
}

/**
 * @brief  桼̾Ѥʸ󥪥֥Ȥ֤
 *
 * @return  ʸ󥪥֥(tstring)λ
 */
tstring &ftpstreamio::username()
{
    return this->username_rec;
}

/**
 * @brief  ѥѤʸ󥪥֥Ȥ֤
 *
 * @return  ʸ󥪥֥(tstring)λ
 */
tstring &ftpstreamio::password()
{
    return this->password_rec;
}

/**
 * @brief  ɸϤΥץ
 *
 *  ɸϤ򥪡ץ󤷤ޤ<br>
 *  mode  "r" ꤵ줿ɸϤ"w" ꤵ줿ɸϤ
 *  ץ󤷤ޤ.
 *
 * @deprecated  Ѥ̣Ϥޤ<br>
 *              ΥдؿǤϡѥꤷơURL򥪡ץ󤹤
 *              Ǥޤ
 * @param     mode ȥ꡼򳫤⡼ ("r" or "w")
 * @return    0<br>
 *            顼ξ
 * @throw     err_rec ɸϤΥե롦ǥץʣ˼Ԥ
 * @note      ͳǺ줿дؿǤ
 */
int ftpstreamio::open( const char *mode )
{
    assert( mode != NULL );

    return this->cstreamio::open(mode);
}

/**
 * @brief  ǥץˤ륹ȥ꡼Υץ
 *
 *  ǥץꤷơȥ꡼򥪡ץ󤷤ޤ
 *
 * @deprecated  Ѥ̣Ϥޤ<br>
 *              ΥдؿǤϡѥꤷơURL򥪡ץ󤹤
 *              Ǥޤ
 * @param       mode ȥ꡼򳫤⡼ ("r" or "w")
 * @param       fd   եǥץ
 * @return      0<br>
 *              顼ξ
 * @note        ͳǺ줿дؿǤ
 */
int ftpstreamio::open( const char *mode, int fd )
{
    assert( mode != NULL );

    int return_status = -1;
    int fd_new = -1;

    if ( this->fp != NULL ) goto quit;

    if ( fd == 0 ) {
        if ( c_strncmp(mode,"r",1) != 0 ) goto quit;
    } else if (fd == 1 || fd == 2) {
        if ( c_strncmp(mode,"w",1) != 0 ) goto quit;
    } else {
        /* DO_NOTHIG */
    }
    fd_new = c_dup(fd);
    this->fp = c_fdopen(fd_new,mode);
    if ( this->fp == NULL ) goto quit;

    return_status = 0;
 quit:
    if ( return_status != 0 && 0 <= fd_new ) c_close(fd_new);
    return return_status;
}

/**
 * @brief  ȥ꡼Υ
 *
 *  open()ǳȥ꡼򥯥ޤ
 *
 * @return     0<br>
 *             顼ξ0ʳ
 */
int ftpstreamio::close()
{
    int return_status = 0;
    if ( this->filter_sp != NULL ) {
	return_status = this->filter_sp->close();
	delete this->filter_sp;
	this->filter_sp = NULL;
    }
    if ( this->rw_sp != NULL ) {
	int status;
	status = this->rw_sp->close();
	delete this->rw_sp;
	this->rw_sp = NULL;
	if ( status != 0 ) return_status = status;
    }
    if ( this->inet_sp != NULL ) {
	tstring ans;
	/* xxx Transfer complete. */
	if ( 400 <= this->read_response(this->inet_sp,ans) ) {
	    err_report1(__FUNCTION__,"ERROR","ftpd: [%s]",ans.strtrim().cstr());
	    return_status = EOF;
	}
	this->inet_sp->printf("QUIT\n");
	this->inet_sp->flush();
	if ( 400 <= this->read_response(this->inet_sp,ans) ) {
	    err_report1(__FUNCTION__,"ERROR","ftpd: [%s]",ans.strtrim().cstr());
	    return_status = EOF;
	}
    }
    if ( this->inet_sp != NULL ) {
	int status;
	status = this->inet_sp->close();
	delete this->inet_sp;
	this->inet_sp = NULL;
	if ( status != 0 ) return_status = status;
    }
    this->content_length_rec = -1;

    if ( this->fp != NULL ) {
	int status;
        status = c_fclose(this->fp);
        this->fp = NULL;
	if ( status != 0 ) return_status = status;
    }

    this->cstreamio::close();

    return return_status;
}

/**
 * @brief  񤭹ѥȥ꡼Υեå
 *
 *  ȥ꡼ˡХåե˳ǼƤ뤹٤ƤΥǡ񤭹ߤޤ
 *
 */
int ftpstreamio::flush()
{
    if ( this->filter_sp != NULL ) {
        return this->filter_sp->flush();
    }
    else if (this->rw_sp != NULL) {
        return this->rw_sp->flush();
    }
    else return this->cstreamio::flush();
}

/**
 * @brief  EOFؼҤμ
 */
int ftpstreamio::eof()
{
    if ( this->filter_sp != NULL ) {
        return this->filter_sp->eof();
    }
    else if (this->rw_sp != NULL) {
        return this->rw_sp->eof();
    }
    else return this->cstreamio::eof();
}

/**
 * @brief  顼ؼҤμ
 */
int ftpstreamio::error()
{
    if ( this->filter_sp != NULL ) {
        return this->filter_sp->error();
    }
    else if (this->rw_sp != NULL) {
        return this->rw_sp->error();
    }
    else return this->cstreamio::error();
}

/**
 * @brief  EOFؼҤȥ顼ؼҤΥꥻå
 */
ftpstreamio &ftpstreamio::reseterr()
{
    if ( this->filter_sp != NULL ) {
        this->filter_sp->reseterr();
    }
    else if (this->rw_sp != NULL) {
        this->rw_sp->reseterr();
    }
    else this->cstreamio::reseterr();
    return *this;
}


/**
 * @brief  ɤ߹ѸϴؿΥå (Υ饹ˤƥХ饤)
 *
 * @param    buf ǡγǼ
 * @param    size ǡθĿ
 * @return   ϥХȿ<br>
 *           ȥ꡼बξ硤0ХȤɤ߹ߤꤵ줿0<br>
 *           顼ξ
 * @note     Υдؿ protected Ǥ
 */
ssize_t ftpstreamio::raw_read( void *buf, size_t size )
{
    if ( this->filter_sp != NULL ) {
	return this->filter_sp->read(buf,size);
    }
    else if ( this->rw_sp != NULL ) {
	return this->rw_sp->read(buf,size);
    }
    else return -1;
}

/**
 * @brief  񤭹ѸϴؿΥå (Υ饹ˤƥХ饤)
 *
 * @param      buf ǡγǼ
 * @param      size ǡθĿ
 * @return     ϥХȿ<br>
 *             顼ξ
 * @note       Υдؿ protected Ǥ
 */
ssize_t ftpstreamio::raw_write( const void *buf, size_t size )
{
    if ( this->filter_sp != NULL ) {
	return this->filter_sp->write(buf,size);
    }
    else if ( this->rw_sp != NULL ) {
	return this->rw_sp->write(buf,size);
    }
    else return -1;
}

}	/* namespace sli */


#include <pwd.h>

/**
 * @brief  ߤΥޥ󤫤桼̾ ()
 *
 * @note   private ʴؿǤ
 */
static void get_user_at_host( sli::tstring *ret )
{
    char host_name[256];
    struct passwd *user_info;
    const char *user_name = NULL;
    /* ۥ̾ */
    if ( c_gethostname(host_name,256) != 0 ) {
	c_strcpy(host_name,"VeryLongHostname");
    }
    host_name[255] = '\0';
    /* 桼̾ */
    user_info = getpwuid(c_geteuid());
    if ( user_info != NULL ) {
	user_name = user_info->pw_name;
    }
    if ( user_name == NULL ) user_name = "unknown";
    /* */
    ret->assign(user_name).append("@").append(host_name);

    return;
}

#include "private/c_gethostname.c"

#include "private/c_fdopen.c"
#include "private/c_fclose.c"

#include "private/c_strcpy.c"
#include "private/c_strncmp.cc"

#include "private/c_geteuid.c"
#include "private/c_dup.c"
#include "private/c_close.c"

#include "private/suffix2filetype.c"
