/* -*- Mode: C++ ; Coding: euc-japan -*- */
/* Time-stamp: <2013-05-12 00:00:46 cyamauch> */

/**
 * @file   tregex.cc
 * @brief  POSIXĥɽ򰷤δŪʥ饹 tregex Υ
 */

#define CLASS_NAME "tregex"

#include "config.h"

#include "tregex.h"

#include <stdlib.h>

#include "private/err_report.h"

#include "private/c_strcmp.h"
#include "private/c_strdup.h"
#include "private/c_regex.h"
#include "private/c_regfatal.h"
#include "private/c_regsearchx.h"
#include "private/memswap.h"

namespace sli
{

/**
 * @brief  󥹥ȥ饯
 *
 */
tregex::tregex()
{
    this->regstr_rec = NULL;
    this->preg_rec = NULL;
    this->errstr_rec[0] = '\0';

    this->shallow_copy_ok = false;
    this->shallow_copy_dest_obj = NULL;
    this->shallow_copy_src_obj = NULL;
    this->__copying = false;

    return;
}

/**
 * @brief  󥹥ȥ饯
 *
 *  ꤵ줿ɽ򥳥ѥ뤷η̤򼫿Ȥ˳Ǽޤ
 * 
 * @param  regstr ѥ뤵ɽʸ
 * @throw  ꤬­Ƥ硥Хåեγݤ˼Ԥ硥
 * @note   ѥ륨顼򸡽Фϡ.compile() ȤǤ
 */
tregex::tregex(const char *regstr)
{
    this->regstr_rec = NULL;
    this->preg_rec = NULL;
    this->errstr_rec[0] = '\0';

    this->shallow_copy_ok = false;
    this->shallow_copy_dest_obj = NULL;
    this->shallow_copy_src_obj = NULL;
    this->__copying = false;

    if ( this->compile(regstr) < 0 ) {
	//err_report(__FUNCTION__,"WARNING","this->compile() reports error");
    }

    return;
}

/**
 * @brief  ԡ󥹥ȥ饯
 *
 *  objƤǼȤޤ
 *
 * @param  obj Ȥʤ륪֥
 */
tregex::tregex(const tregex &obj)
{
    this->regstr_rec = NULL;
    this->preg_rec = NULL;
    this->errstr_rec[0] = '\0';

    this->shallow_copy_ok = false;
    this->shallow_copy_dest_obj = NULL;
    this->shallow_copy_src_obj = NULL;
    this->__copying = false;

    this->init(obj);

    return;
}

/**
 * @brief  ǥȥ饯
 * 
 */
tregex::~tregex()
{
    if ( this->regstr_rec != NULL ) {
	free(this->regstr_rec);
    }
    if ( this->preg_rec != NULL ) {
	c_regfree(this->preg_rec);
    }
    return;
}

/**
 * @brief  ֥ȤΥԡ
 *
 *  黻Ҥα¦() ǻꤵ줿 tregex ֥Ȥ򼫿Ȥ˥ԡޤ
 *
 * @param   obj Ȥʤ륪֥
 * @return  Ȥλ
 */
tregex &tregex::operator=(const tregex &obj)
{
    this->init(obj);
    return *this;
}


/**
 * @brief  ꤵ줿ɽ򥳥ѥ뤷Ȥ˳Ǽ
 *
 *  黻Ҥα¦() ǻꤵ줿ɽ򥳥ѥ뤷η̤򼫿Ȥ
 *  Ǽޤ
 *
 * @param  regstr ѥ뤵ɽʸ
 * @throw  ꤬­Ƥ硥Хåեγݤ˼Ԥ硥
 * @note   ѥ륨顼򸡽Фϡ.compile() ȤǤ
 */
tregex &tregex::operator=(const char *regstr)
{
    if ( this->compile(regstr) < 0 ) {
	//err_report(__FUNCTION__,"WARNING","this->compile() reports error");
    }
    return *this;
}

/**
 * @brief   ֥Ȥν
 * 
 * @return  Ȥλ
 */
tregex &tregex::init()
{
    if ( this->regstr_rec != NULL ) {
	free(this->regstr_rec);
	this->regstr_rec = NULL;
    }
    if ( this->preg_rec != NULL ) {
	c_regfree(this->preg_rec);
	this->preg_rec = NULL;
    }
    this->errstr_rec[0] = '\0';
    return *this;
}

/**
 * @brief  ֥ȤΥԡ
 *
 *  obj Ƥ򼫿Ȥ˥ԡޤ
 *
 * @param   obj tregex֥
 * @return  Ȥλ
 * @throw   ꤬­Ƥ硥Хåեγݤ˼Ԥ硥
 */
tregex &tregex::init(const tregex &obj)
{
    if ( &obj == this ) return *this;
    this->tregex::init();
    if ( obj.cstr() != NULL ) this->tregex::compile(obj.cstr());
    return *this;
}

/* 顼å cregex() Ǥǽcregex() ֤ͤ NULL ʤ */
/* ѥ˼ԤƤ롥                                        */

/**
 * @brief  ꤵ줿ɽ򥳥ѥ뤷 ̤򼫿Ȥ˳Ǽ (֤ͤstatus)
 *
 *  ɽ regstr 򥳥ѥ뤷ɽȥѥ̤Ȥ򥪥֥Ȥ
 *  Ǽޤ
 * 
 * @param   regstr ѥ뤵ɽʸ
 * @return  0: ｪλ <br>
 *          : ɽʸʾ
 * @throw   ꤬­Ƥ硥Хåեγݤ˼Ԥ硥
 */
int tregex::compile(const char *regstr)
{
    int status;
    void *tmp_preg;

    if ( this->regstr_rec != NULL && regstr != NULL &&
	 c_strcmp(this->regstr_rec,regstr) == 0 ) return 0;	/* OK */

    if ( this->regstr_rec != NULL ) {
	free(this->regstr_rec);
	this->regstr_rec = NULL;
    }
    if ( this->preg_rec != NULL ) {
	c_regfree(this->preg_rec);
	this->preg_rec = NULL;
    }
    this->errstr_rec[0] = '\0';

    if ( regstr == NULL ) return -1;				/* Invalid */

    status = c_regcomp(&tmp_preg, regstr);
    if ( status != 0 ) {
	if ( c_regfatal(status) ) {
	    c_regfree(tmp_preg);
	    err_throw(__FUNCTION__,"FATAL","Out of memory in c_regcomp()");
	}
	c_regerror(status,tmp_preg,this->errstr_rec,192);
	c_regfree(tmp_preg);
	return -1;						/* Error */
    }

    /* register */
    this->preg_rec = tmp_preg;
    this->regstr_rec = c_strdup(regstr);
    if ( this->regstr_rec == NULL ) {
	c_regfree(this->preg_rec);
	this->preg_rec = NULL;
	err_throw(__FUNCTION__,"FATAL","c_strdup() failed");
    }

    return 0;
}

/**
 * @brief  2ĤΥ֥ȴ֤ǤƤΥå
 *
 *  ȤƤ sobj ƤȤؤޤ
 *
 * @param   sobj tregex ֥
 * @return  Ȥλ
 */
tregex &tregex::swap(tregex &sobj)
{
    if ( &sobj == this ) return *this;

    char *tmp__regstr_rec;
    void *tmp__preg_rec;

    tmp__regstr_rec = sobj.regstr_rec;
    tmp__preg_rec  = sobj.preg_rec;
    sobj.regstr_rec = this->regstr_rec;
    sobj.preg_rec  = this->preg_rec;
    this->regstr_rec = tmp__regstr_rec;
    this->preg_rec  = tmp__preg_rec;

    memswap(this->errstr_rec, sobj.errstr_rec, 192);

    return *this;
}

/**
 * @brief  Ȥ˳ǼƤɽʸ֤
 *
 * @return  ǼƤɽʸ
 */
const char *tregex::cstr() const
{
    return this->regstr_rec;
}

/**
 * @brief  Ȥ˳ǼƤ regex_t ֥ȤΥɥ쥹֤
 * 
 *  Ȥ˳ǼƤ regex_t ֥ȤΥɥ쥹֤ޤ<br>
 *  Υдؿ NULL ֤硤compile() ˼ԤƤ򼨤ޤ
 *
 * @return  Ȥ˳ǼƤ regex_t ֥ȤΥɥ쥹
 */
const void *tregex::cregex() const
{
    return this->preg_rec;
}

/**
 * @brief  ꤵ줿ʸˤĤơɽˤޥå󥰤
 *
 *  ʸ str ˤĤơȤ˳Ǽ줿ɽˤޥå󥰤Ԥޤ
 * 
 * @param  str оʸ
 * @param  notbol true ξ硤ƬޥåΥڥ졼ɬԤ
 * @param  noteol true ξ硤ޥåΥڥ졼ɬԤ
 * @param  max_nelem ¤θĿ
 * @param  pos_r[] ޥå֤Ǽ
 * @param  len_r[] ޥåʸǿǼ
 * @param  nelem_r ޥåĿǼ
 * @return  0: ޥå <br>
 *          0ʳ: ޥå󥰼
 */
int tregex::exec( const char *str, bool notbol, bool noteol,
		  size_t max_nelem, size_t pos_r[], size_t len_r[],
		  size_t *nelem_r ) const
{
    return c_regsearchx(this->cregex(), str, notbol, noteol,
			max_nelem, pos_r, len_r, nelem_r);
}

/**
 * @brief  ѥΥ顼å
 *
 *  Ȥ˳ǼƤ ѥΥ顼å֤ޤ<br>
 *  줬 NULL ֤硤compile() Ƥ򼨤ޤ
 *
 * @return  ǼƤ륨顼å: ѥ˼Ԥ <br>
 *          NULL: ѥ
 */
const char *tregex::cerrstr() const
{
    if ( this->preg_rec != NULL ) return NULL;
    else return this->errstr_rec;
}

/**
 * @brief  ƥݥꥪ֥ȤΤshallow copy°Ϳ (̤)
 * @deprecated  ̤
 *
 */
/* : Ȥ֤Ȥreturn ʸǤȥ󥹥 */
/*       2󥳥ԡȤʻˤʤΤ void ˤʤäƤ */
void tregex::set_scopy_flag()
{
    this->shallow_copy_ok = true;
    return;
}

/**
 * @brief  shallow copy ǽ֤ (̤)
 * 
 * @return  shallow copyǽʤ鿿<br>
 *          ʳλϵ
 * @note    Υдؿ private Ǥ
 */
bool tregex::request_shallow_copy( tregex *from_obj ) const
{
    return false;
}


}

#include "private/c_strcmp.cc"
#include "private/c_strdup.c"
#include "private/c_regex.c"
#include "private/c_regfatal.c"
#include "private/c_regsearchx.c"
#include "private/memswap.cc"

